"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.secureMask = exports.mask = exports.cardinalityBits = exports.checkPassword = exports.alphabet = exports.product = exports.and = exports.or = exports.formatDuration = exports.DATE = exports.zip = void 0;
const hexes = Array.from({ length: 256 }, (v, i) => i.toString(16).padStart(2, '0'));
function bytesToHex(uint8a) {
    if (!(uint8a instanceof Uint8Array))
        throw new Error('Uint8Array expected');
    let hex = '';
    for (let i = 0; i < uint8a.length; i++) {
        hex += hexes[uint8a[i]];
    }
    return hex;
}
function numberToHexUnpadded(num) {
    let hex = num.toString(16);
    hex = hex.length & 1 ? `0${hex}` : hex;
    return hex;
}
function strip0x(hex) {
    return hex.replace(/^0x/i, '');
}
function hexToBytes(hex) {
    hex = strip0x(hex);
    hex = hex.length & 1 ? `0${hex}` : hex;
    if (typeof hex !== 'string') {
        throw new TypeError('hexToBytes: expected string, got ' + typeof hex);
    }
    if (hex.length % 2)
        throw new Error('hexToBytes: received invalid unpadded hex');
    const array = new Uint8Array(hex.length / 2);
    for (let i = 0; i < array.length; i++) {
        const j = i * 2;
        const hexByte = hex.slice(j, j + 2);
        const byte = Number.parseInt(hexByte, 16);
        if (Number.isNaN(byte) || byte < 0)
            throw new Error('Invalid byte sequence');
        array[i] = byte;
    }
    return array;
}
function bytesToNumber(bytes) {
    return BigInt('0x' + bytesToHex(bytes));
}
function zip(a, b) {
    let res = [];
    for (let i = 0; i < Math.max(a.length, b.length); i++)
        res.push([a[i], b[i]]);
    return res;
}
exports.zip = zip;
exports.DATE = { sec: 1000 };
exports.DATE.min = 60 * exports.DATE.sec;
exports.DATE.h = 60 * exports.DATE.min;
exports.DATE.d = 24 * exports.DATE.h;
exports.DATE.mo = 30 * exports.DATE.d;
exports.DATE.y = 365 * exports.DATE.mo;
function formatDuration(dur) {
    if (Number.isNaN(dur))
        return 'never';
    if (dur > exports.DATE.y * 100)
        return 'centuries';
    let parts = [];
    for (let [name, period] of Object.entries(exports.DATE).reverse()) {
        if (dur < period)
            continue;
        let value = Math.floor(dur / period);
        parts.push(`${value}${name}`);
        dur -= value * period;
    }
    return !parts.length ? '0 sec' : parts.join(' ');
}
exports.formatDuration = formatDuration;
function or(...sets) {
    return sets.reduce((acc, i) => new Set([...acc, ...i]), new Set());
}
exports.or = or;
function and(...sets) {
    return sets.reduce((acc, i) => new Set(Array.from(acc).filter((j) => i.has(j))));
}
exports.and = and;
function product(...sets) {
    return sets.reduce((acc, i) => new Set(Array.from(acc)
        .map((j) => Array.from(i).map((k) => j + k))
        .flat()));
}
exports.product = product;
exports.alphabet = {};
exports.alphabet['1'] = new Set('0123456789');
exports.alphabet['@'] = new Set('!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~');
exports.alphabet['v'] = new Set('aeiouy');
exports.alphabet['c'] = new Set('bcdfghjklmnpqrstvwxz');
exports.alphabet['a'] = or(exports.alphabet['v'], exports.alphabet['c']);
for (const v of 'vca')
    exports.alphabet[v.toUpperCase()] = new Set(Array.from(exports.alphabet[v]).map((i) => i.toUpperCase()));
exports.alphabet['l'] = or(exports.alphabet['a'], exports.alphabet['A']);
exports.alphabet['n'] = or(exports.alphabet['l'], exports.alphabet['1']);
exports.alphabet['*'] = or(exports.alphabet['n'], exports.alphabet['@']);
const TEMPLATES = {
    s: 'cv',
    S: 'Cv',
};
function idx(arr, i) {
    if (!Array.isArray(arr))
        arr = Array.from(arr);
    if (i < 0 || i >= arr.length)
        throw new Error('Out of bounds index access');
    return arr[i];
}
function checkPassword(pwd) {
    if (pwd.length < 8)
        return false;
    const s = new Set(pwd);
    for (const c of 'aA1@')
        if (!and(s, exports.alphabet[c]).size)
            return false;
    return true;
}
exports.checkPassword = checkPassword;
function splitEntropy(lengths, entropy) {
    let entropyLeft = bytesToNumber(entropy);
    let values = [];
    for (const c of lengths) {
        const sz = BigInt(c);
        values.push(Number(entropyLeft % sz));
        entropyLeft /= sz;
    }
    return { values, entropyLeft };
}
function cardinalityBits(cardinality) {
    let i = 0;
    for (let c = cardinality; c; i++, c >>= 1n)
        ;
    return i - 1;
}
exports.cardinalityBits = cardinalityBits;
const guessTime = (cardinality, perSec) => formatDuration((Number(cardinality) / perSec) * 1000);
function passwordScore(cardinality) {
    const scores = [
        [1e3 + 5, 'too guessable'],
        [1e6 + 5, 'very guessable'],
        [1e8 + 5, 'somewhat guessable'],
        [1e10 + 5, 'safely unguessable'],
    ];
    let res = 'very unguessable';
    for (const [i, v] of scores) {
        if (cardinality <= BigInt(i)) {
            res = v;
            break;
        }
    }
    return res;
}
function estimateAttack(cardinality) {
    const GPU_PRICE = 20500 / 6;
    const GPU_COST = GPU_PRICE / (2 * (exports.DATE.y / 1000));
    const GPU_POWER = 320;
    const GPU_POWER_RIG = (80 + 280 + 6 * GPU_POWER) / 6;
    const KWH_PRICE = 0.12;
    const KWH_COOLING = KWH_PRICE + KWH_PRICE * 0.33;
    const WS = KWH_COOLING / 60 / 1000;
    const ENERGY_COST = GPU_POWER_RIG * WS;
    const TOTAL_GPU_COST = ENERGY_COST + GPU_COST;
    const calcCost = (hashes) => Number(cardinality / BigInt(hashes)) * TOTAL_GPU_COST;
    return {
        score: passwordScore(cardinality),
        guesses: {
            online_throttling: guessTime(cardinality, 100 / (exports.DATE.h / 1000)),
            online: guessTime(cardinality, 10),
            slow: guessTime(cardinality, 10000),
            fast: guessTime(cardinality, 10000000000),
        },
        costs: {
            luks: calcCost(22779),
            filevault2: calcCost(151300),
            macos: calcCost(1019200),
            pbkdf2: calcCost(3029200),
        },
    };
}
class Mask {
    constructor(mask) {
        mask = mask
            .split('')
            .map((i) => TEMPLATES[i] || i)
            .join('');
        this.chars = mask.split('');
        this.length = this.chars.length;
        this.sets = this.chars.map((i) => exports.alphabet[i] || new Set([i]));
        this.lengths = this.sets.map((i) => i.size);
        this.cardinality = this.sets.reduce((acc, i) => acc * BigInt(i.size), 1n);
        this.entropy = cardinalityBits(this.cardinality);
    }
    apply(entropy) {
        if (this.cardinality >= 2n ** BigInt((8 * entropy.length) / 2))
            throw new Error('Not enough entropy');
        const { entropyLeft, values } = splitEntropy(this.lengths, entropy);
        const password = zip(this.sets, values)
            .map(([s, v]) => idx(s, v))
            .join('');
        return { password, entropyLeft };
    }
    inverse({ password, entropyLeft }) {
        const values = zip(this.sets, password.split('')).map(([s, c]) => Array.from(s).indexOf(c));
        const num = zip(this.sets, values).reduceRight((acc, [s, v]) => acc * BigInt(s.size) + BigInt(v), 0n);
        return hexToBytes(numberToHexUnpadded(entropyLeft * this.cardinality + num));
    }
    estimate() {
        return estimateAttack(this.cardinality);
    }
}
const mask = (mask) => new Mask(mask);
exports.mask = mask;
const secureMasks = [];
for (let upper = 0; upper < 17; upper++) {
    for (let digitPos = 0; digitPos < 3; digitPos++) {
        for (let digitLeft = 0; digitLeft < 2; digitLeft++) {
            const groups = ['cvccvc', 'cvccvc', 'cvccvc'];
            groups[digitPos] = digitLeft ? '1cvcvc' : 'cvccv1';
            const mask = groups.join('-');
            let res;
            for (let i = 0, sI = 0; i < mask.length; i++) {
                const chr = mask[i];
                if (!['c', 'v'].includes(chr))
                    continue;
                if (sI === upper)
                    res = mask.slice(0, i) + chr.toUpperCase() + mask.slice(i + 1);
                sI++;
            }
            if (!res)
                throw new Error('Cannot find uppercase syllable index');
            secureMasks.push(res);
        }
    }
}
exports.secureMask = (() => {
    const size = BigInt(secureMasks.length);
    const cardinality = (0, exports.mask)(secureMasks[0]).cardinality * size;
    return {
        length: 20,
        cardinality,
        entropy: cardinalityBits(cardinality),
        estimate: () => estimateAttack(cardinality),
        apply: (entropy) => {
            let entropyLeft = bytesToNumber(entropy);
            const idx = Number(entropyLeft % size);
            return (0, exports.mask)(secureMasks[idx]).apply(hexToBytes(numberToHexUnpadded(entropyLeft / size)));
        },
        inverse(res) {
            const chars = res.password.split('');
            const maskStr = chars
                .map((i) => {
                const possibleValues = Object.entries(exports.alphabet)
                    .filter(([c, _]) => ['c', 'v', 'C', 'V', '1'].includes(c))
                    .map(([c, v]) => [c, and(v, new Set([i]))])
                    .filter(([_, v]) => v.size);
                if (possibleValues.length > 1)
                    throw new Error('Too much possible values, cannot detect mask.');
                return possibleValues.length ? possibleValues[0][0] : i;
            })
                .join('');
            const idx = secureMasks.indexOf(maskStr);
            if (idx < 0)
                throw new Error('Unknown mask');
            const entropy = (0, exports.mask)(secureMasks[idx]).inverse(res);
            const entropyNum = bytesToNumber(entropy);
            return hexToBytes(numberToHexUnpadded(entropyNum * size + BigInt(idx)));
        },
    };
})();
